from PIL import Image
import operator
from collections import deque
from io import StringIO

#adapted from: https://github.com/IngJavierR/PngToSvg
# and https://github.com/ianmackinnon/png2svg

def add_tuple(a, b):
    return tuple(map(operator.add, a, b))

def sub_tuple(a, b):
    return tuple(map(operator.sub, a, b))

def neg_tuple(a):
    return tuple(map(operator.neg, a))

def direction(edge):
    return sub_tuple(edge[1], edge[0])

def magnitude(a):
    return int(pow(pow(a[0], 2) + pow(a[1], 2), .5))

def normalize(a):
    mag = magnitude(a)
    assert mag > 0, "Cannot normalize a zero-length vector"
    return tuple(map(operator.truediv, a, [mag]*len(a)))

def svg_header(width, height):
    return """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"
  "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">
<svg width="%d" height="%d"
     xmlns="http://www.w3.org/2000/svg" version="1.1">
""" % (width, height)

def joined_edges(assorted_edges, keep_every_point=False):
    pieces = []
    piece = []
    directions = deque([
        (0, 1),
        (1, 0),
        (0, -1),
        (-1, 0),
        ])
    while assorted_edges:
        if not piece:
            piece.append(assorted_edges.pop())
        current_direction = normalize(direction(piece[-1]))
        while current_direction != directions[2]:
            directions.rotate()
        for i in range(1, 4):
            next_end = add_tuple(piece[-1][1], directions[i])
            next_edge = (piece[-1][1], next_end)
            if next_edge in assorted_edges:
                assorted_edges.remove(next_edge)
                if i == 2 and not keep_every_point:
                    # same direction
                    piece[-1] = (piece[-1][0], next_edge[1])
                else:
                    piece.append(next_edge)
                if piece[0][0] == piece[-1][1]:
                    if not keep_every_point and normalize(direction(piece[0])) == normalize(direction(piece[-1])):
                        piece[-1] = (piece[-1][0], piece.pop(0)[1])
                        # same direction
                    pieces.append(piece)
                    piece = []
                break
        else:
            raise Exception ("Failed to find connecting edge")
    return pieces

def rgba_image_to_svg_contiguous(im, opaque=None, keep_every_point=False):

    # collect contiguous pixel groups
    
    adjacent = ((1, 0), (0, 1), (-1, 0), (0, -1))
    visited = Image.new("1", im.size, 0)
    
    color_pixel_lists = {}

    width, height = im.size
    for x in range(width):
        for y in range(height):
            here = (x, y)
            if visited.getpixel(here):
                continue
            rgba = im.getpixel((x, y))
            if opaque and not rgba[3]:
                continue
            piece = []
            queue = [here]
            visited.putpixel(here, 1)
            while queue:
                here = queue.pop()
                for offset in adjacent:
                    neighbour = add_tuple(here, offset)
                    if not (0 <= neighbour[0] < width) or not (0 <= neighbour[1] < height):
                        continue
                    if visited.getpixel(neighbour):
                        continue
                    neighbour_rgba = im.getpixel(neighbour)
                    if neighbour_rgba != rgba:
                        continue
                    queue.append(neighbour)
                    visited.putpixel(neighbour, 1)
                piece.append(here)

            if not rgba in color_pixel_lists:
                color_pixel_lists[rgba] = []
            color_pixel_lists[rgba].append(piece)

    del adjacent
    del visited

    # calculate clockwise edges of pixel groups

    edges = {
        (-1, 0):((0, 0), (0, 1)),
        (0, 1):((0, 1), (1, 1)),
        (1, 0):((1, 1), (1, 0)),
        (0, -1):((1, 0), (0, 0)),
        }
            
    color_edge_lists = {}

    for rgba, pieces in color_pixel_lists.items():
        for piece_pixel_list in pieces:
            edge_set = set([])
            for coord in piece_pixel_list:
                for offset, (start_offset, end_offset) in edges.items():
                    neighbour = add_tuple(coord, offset)
                    start = add_tuple(coord, start_offset)
                    end = add_tuple(coord, end_offset)
                    edge = (start, end)
                    if neighbour in piece_pixel_list:
                        continue
                    edge_set.add(edge)
            if not rgba in color_edge_lists:
                color_edge_lists[rgba] = []
            color_edge_lists[rgba].append(edge_set)

    del color_pixel_lists
    del edges

    # join edges of pixel groups

    color_joined_pieces = {}

    for color, pieces in color_edge_lists.items():
        color_joined_pieces[color] = []
        for assorted_edges in pieces:
            color_joined_pieces[color].append(joined_edges(assorted_edges, keep_every_point))

    s = StringIO()
    s.write(svg_header(*im.size))

    for color, shapes in color_joined_pieces.items():
        for shape in shapes:
            s.write(""" <path d=" """)
            for sub_shape in shape:
                here = sub_shape.pop(0)[0]
                s.write(""" M %d,%d """ % here)
                for edge in sub_shape:
                    here = edge[0]
                    s.write(""" L %d,%d """ % here)
                s.write(""" Z """)
            s.write(""" " style="fill:rgb%s; fill-opacity:%.3f; stroke:none;" />\n""" % (color[0:3], float(color[3]) / 255))
            
    s.write("""</svg>\n""")
    return s.getvalue()

def rgba_image_to_svg_pixels(im, opaque=None):
    s = StringIO()
    s.write(svg_header(*im.size))

    width, height = im.size
    for x in range(width):
        for y in range(height):
            here = (x, y)
            rgba = im.getpixel(here)
            if opaque and not rgba[3]:
                continue
            s.write("""  <rect x="%d" y="%d" width="1" height="1" style="fill:rgb%s; fill-opacity:%.3f; stroke:none;" />\n""" % (x, y, rgba[0:3], float(rgba[2]) / 255))
    s.write("""</svg>\n""")
    return s.getvalue()

def convert_to_svg(src,dest):
    image = Image.open(src).convert('RGBA')
    svg_image = rgba_image_to_svg_contiguous(image)
    with open(dest,"w") as text_file:
        text_file.write(svg_image)



import os
from os import path
import sys
import unittest
import warnings
import pyemap
import tempfile
from sys import platform

def test_save_functions():
    my_emap = pyemap.parse(os.path.join(sys.path[0],"pyemap/tests/test_pdbs/4DJA.pdb")) 
    #cluster
    fout = tempfile.NamedTemporaryFile(suffix=".png")
    fout_svg = tempfile.NamedTemporaryFile(suffix=".svg")
    #aromatic eta moiety
    my_emap.residue_to_Image("SF4603(A)")
    my_emap.residue_to_file("SF4603(A)",dest=fout.name)
    my_emap.residue_to_Image("FAD601(A)-1")
    my_emap.residue_to_file("FAD601(A)-1",dest=fout.name)
    my_emap.residue_to_file("FAD601(A)-1",dest=fout_svg.name)
    if platform == "linux":
        pyemap.process(my_emap,sdef=1)
    elif platform == "darwin":
        pyemap.process(my_emap)
    #standard residue
    my_emap.residue_to_Image("Y443(A)")
    my_emap.residue_to_file("Y443(A)",dest=fout.name)
    #init graph
    my_emap.init_graph_to_Image()
    my_emap.init_graph_to_file(dest=fout.name)
    pyemap.find_paths(my_emap,"Y443(A)",target="Y437(A)")
    #paths graph
    my_emap.paths_graph_to_Image()
    my_emap.paths_graph_to_file(dest=fout.name)
    #check that report does something
    assert my_emap.report() != None
    



# PyeMap: A python package for automatic identification of electron and hole transfer pathways in proteins.
# Copyright(C) 2017-2020 Ruslan Tazhigulov, James Gayvert, Ksenia Bravaya (Boston University, USA)
from .custom_residues import is_pi_bonded
from .data import clusters
import networkx as nx
from networkx.drawing.nx_agraph import to_agraph
from .structures import getSimpleSmarts
from collections import OrderedDict
from PIL import Image
import os
from shutil import copyfile
import tempfile
from svglib.svglib import svg2rlg
from reportlab.graphics import renderPM
from .process_data import get_atom_list
from .png2svg import convert_to_svg


class emap():
    '''
    Manages the data generated at all stages of PyeMap analysis.

    Attributes
    ----------
    filename: str
        Name of the crystal structure file being processed by PyeMap.
    structure: :class:`Bio.PDB.Structure.Structure`
        Macromolecular protein structure. Contains model which contains residues and atoms.
    eta_moieties: dict of node label(str): :class:`Bio.PDB.Residue.Residue`
        Non-protein eta moieties automatically identified at the parsing step.
    chain_list: list of str
        List of chains identified at the parsing step.
    smarts: dict of node label(str): smarts string(str)
        List of SMARTS strings for non-protein eta moieties identified at the parsing step.
    residues: dict of node label(str): :class:`Bio.PDB.Residue.Residue`
        Residues included in the graph after the process step.
    ngl_strings: dict of nodel label(str):ngl_string(str)
        Formatted NGL viewer selection strings for residues included in the graph after the process step.
    user_residues: dict of str: :class:`Bio.PDB.Residue.Residue`
        Custom residues specified by the user.
    init_graph: :class:`networkx.Graph`
        Graph generated after the process step.
    branches: dict of branch id(int): :class:`~pyemap.Branch`
        Branches found by PyeMap analysis
    paths: dict of path id(str): :class:`~pyemap.ShortestPath`
        Paths found by PyeMap sorted by lowest to highest score.
    paths_graph: :class:`networkx.Graph`
        Graph generated after the shortest paths step.
    '''

    def __init__(self, filename, structure, eta_moieties, chain_list):
        '''Initializes emap object.

        Parameters
        ----------
        filename: str
            Name of file
        structure: :class:`Bio.PDB.Structure.Structure`
            Macromolecular protein structure
        eta_moieties: list of :class:`Bio.PDB.Residue.Residue`
            Customized residue objects generated for automatically detected eta moieties
        chain_list: list of str
            Chains identified by the parser
        '''
        self.filename = filename
        self.pdb_id = filename[-8:-4].upper()
        self.structure = structure
        self.residues = {}
        self.chains = chain_list
        self.eta_moieties = {}
        self.user_residues = {}
        self.smarts = {}
        self.paths = OrderedDict()
        self.paths_graph = []
        self.init_graph = []
        self.ngl_strings = {}
        self.branches = OrderedDict()
        for residue in eta_moieties:
            self._add_eta_moiety(residue)

    def _store_initial_graph(self, graph):
        '''Stores graph representation of emap model

        Parameters
        ----------
        graph: networkx.Graph
            Graph theory representation of emap model

        Notes
        -----
        Attributes of nodes/edges store info on surface exposure, edge weights etc.
        '''
        self.init_graph = graph

    def _store_paths(self, branches, yens=False):
        '''Stores pathways in emap object, and sets their ngl selection strings for visualization.

        Parameters
        ---------
        shortest_paths: list of pyemap.shortest_paths.Branch
            branches found by emap
        yens: boolean, optional
            True when target specified, False when only source is specified
        '''
        shortest_paths = []
        for br in branches:
            self.branches[br.branch_id] = br
            shortest_paths += br.paths
        shortest_paths = sorted(shortest_paths)
        for pt in shortest_paths:
            self.paths[pt.path_id] = pt
            self._visualize_pathway(pt, yens)

    def _store_paths_graph(self, graph):
        '''Stores graph representation of emap model with selected pathway(s) highlighted

        Parameters
        ----------
        graph: networkx.Graph
            Graph theory representation of emap model

        Notes
        -----
        Attributes of nodes/edges store info on surface exposure, edge weights etc.
        '''
        self.paths_graph = graph

    def _reset_process(self):
        '''Returns emap object to state it was in after parsing.
        '''
        self.residues = {}
        self.user_residues = {}
        self.init_graph = []
        self.paths = OrderedDict()
        self.paths_graph = []
        self.branches = OrderedDict()
        self.ngl_strings = {}

    def _reset_paths(self):
        '''Returns emap object to state it was in after the process step.
        '''
        self.paths = OrderedDict()
        self.paths_graph = []
        self.branches = OrderedDict()

    def _get_residue_graph(self, residue):
        atoms = get_atom_list(residue)
        arom_atoms = ['O', 'P', 'N', 'C', 'S']
        res_graph = nx.Graph()
        for i in range(len(atoms)):
            for k in range(i, len(atoms)):
                if (not i == k) and is_pi_bonded(atoms[i], atoms[k]):
                    if atoms[i].element in arom_atoms and atoms[k].element in arom_atoms:
                        res_graph.add_edge(i, k)
                        res_graph.nodes[i]["element"] = atoms[i].element
                        res_graph.nodes[i]["coords"] = atoms[i].coord
                        res_graph.nodes[k]["element"] = atoms[k].element
                        res_graph.nodes[k]["coords"] = atoms[k].coord
        return res_graph

    def _add_eta_moiety(self, residue):
        '''Gets the SMARTS string for an automatically identified non-protein eta moiety,
        and adds the residue to the eta_moieties dictionary.

        Parameters
        ----------
        residue: Bio.PDB.Residue.Residue
             Customized residue object generated for automatically detected eta moiety
        '''
        if not residue.resname[:3] in clusters and "CUST" not in residue.resname:
            res_graph = self._get_residue_graph(residue)
            smarts_str = getSimpleSmarts(res_graph)
            residue.smarts = smarts_str
            self.smarts[residue.resname] = smarts_str
        self.eta_moieties[residue.resname] = residue

    def _visualize_pathway(self, pathway, yens):
        '''
        '''
        colors = {"F": "orange", "Y": "blue", "W": "red", "H": "green"}
        selection_strs = []
        color_list = []
        labeled_atoms = []
        label_texts = []
        for res in pathway.get_path_as_list()[1:-1]:
            label_texts.append(res)
            try:
                if res not in self.eta_moieties:
                    color_list.append(colors[res[0]])
                    labeled_atoms.append(".CA")
                else:
                    color_list.append("pink")
                    labeled_atoms.append(next(self.residues[res].get_atoms()).name)
            except KeyError:
                color_list.append("pink")
                labeled_atoms.append(next(self.residues[res].get_atoms()).name)
            selection_strs.append(self.residues[res].ngl_string)
        color_list[0] = "yellow"
        if yens:
            color_list[-1] = "turquoise"
        pathway.set_visualization(
            selection_strs, color_list, labeled_atoms, label_texts)

    def _add_residue(self, residue):
        '''Gets ngl string for residue, and adds the residue to the residues and ngl_strings dictionaries.
        '''
        if not residue.resname[:3] in clusters and "CUST" not in residue.resname and residue.resname[:3] in ["PHE","HIS","TYR","TRP"]:
            res_graph = self._get_residue_graph(residue)
            smarts_str = getSimpleSmarts(res_graph)
            residue.Smarts = smarts_str
            self.smarts[residue.node_label] = smarts_str
        residue.ngl_string = self._get_ngl_string(residue)
        self.residues[residue.node_label] = residue
        self.ngl_strings[residue.node_label] = residue.ngl_string

    def _get_ngl_string(self, residue):
        """Returns NGL selection string for residue

        Parameters
        ----------
        residue: Bio.PDB.Residue.Residue object

        Returns
        -------
        select_string: str
            NGL selection string for residue
        """
        select_string = ""
        atm_list = list(residue.get_atoms())
        first_atm = atm_list[0]
        if hasattr(first_atm,"original_id") and first_atm.original_id is not None:
            id = first_atm.original_id
        else:
            id = first_atm.full_id
        select_string += "(" + str(id[3][1]) + " and :" + str(
            id[2]) + " and ^" + residue.id[2] + " and ." + first_atm.name + ")"
        for i in range(1, len(atm_list)):
            atm = atm_list[i]
            if hasattr(atm,"original_id") and atm.original_id is not None:
                id = first_atm.original_id
            else:
                id = atm.full_id
            select_string += " or "
            select_string += "(" + str(id[3][1]) + " and :" + str(
                id[2]) + " and ^" + residue.id[2] + " and ." + atm.name + ")"
        return select_string

    def residue_to_file(self, resname, dest="", size=(100, 100)):
        '''Saves image of residue to file in .svg format.

        Parameters
        ----------
        resname: str
            Name of residue (node label) to be saved to file.
        dest: str, optional
            destination to save the image
        size: (float,float), optional
            dimensions of image saved to file
        '''
        try:
            from rdkit import Chem
            from rdkit.Chem import Draw
        except (ModuleNotFoundError, ImportError) as e:
            raise ModuleNotFoundError(
                "RDKit is required for visualization of chemical structures. See https://www.rdkit.org/docs/Install.html"
            ) from e
        if resname in self.residues or resname in self.eta_moieties:
            if "CUST" in resname:
                raise KeyError("Not available for user defined residues.")
            elif resname[:3] in clusters:
                cluster_img_name = os.path.abspath(os.path.dirname(
                    __file__)) + '/data/clusters/' + resname[:3] + '.svg'
                if dest:
                    target_name = dest
                else:
                    target_name = resname + ".svg"
                copyfile(cluster_img_name, target_name)
            else:
                try:
                    mol = Chem.MolFromSmarts(self.smarts[resname])
                    mol.UpdatePropertyCache()
                    if dest:
                        Draw.MolToFile(mol, dest, kekulize=False, size=size)
                    else:
                        Draw.MolToFile(mol, resname + ".svg",
                                    kekulize=False, size=size)
                except Exception as e:
                    print("Warning: RDKIT couldn't draw: " + dest)
                    print("Trying to convert png to svg...")
                    try:
                        if dest[-4:]==".svg":
                            Draw.MolToFile(mol, dest[:-4]+".png", kekulize=False, size=size)
                            convert_to_svg(dest[:-4]+".png",dest)
                    except Exception as e:
                        print(e)
                        print("Warning: couldn't convert png to svg.")
                            
        else:
            raise KeyError("No record of any residue by that name.")

    def residue_to_Image(self, resname, size=(100, 100)):
        '''Returns PIL image of chemical structure

        Parameters
        -----------
        resname: str
            Name of residue
        size: (float,float), optional
            dimensions of image saved to file
        Returns
        --------
        img: :class:`PIL.Image.Image`
        '''
        try:
            from rdkit import Chem
            from rdkit.Chem import Draw
        except (ModuleNotFoundError, ImportError) as e:
            raise ModuleNotFoundError(
                "RDKit is required for visualization of chemical structures. See https://www.rdkit.org/docs/Install.html"
            ) from e
        if resname in self.residues or resname in self.eta_moieties:
            if "CUST" in resname:
                raise KeyError("Not available for user defined residues.")
            elif resname[:3] in clusters:
                cluster_img_name = os.path.abspath(os.path.dirname(
                    __file__)) + '/data/clusters/' + resname[:3] + '.svg'
                drawing = svg2rlg(cluster_img_name)
                img = renderPM.drawToPIL(drawing)
                return img
            else:
                try:
                    mol = Chem.MolFromSmarts(self.smarts[resname])
                    mol.UpdatePropertyCache()
                    img = Draw.MolToImage(mol, kekulize=False, size=size)
                    return img
                except Exception as e:
                    print("Warning: RDKIT couldn't draw: " + dest)
        else:
            raise KeyError("No record of any residue by that name.")

    def _graph_to_file(self, G, dest=""):
        '''Saves image of graph generated by process step to file.

        Parameters
        ----------
        dest: str
            Destination for writing to file.
        '''
        if G:
            agraph = to_agraph(G)
            agraph.graph_attr.update(
                ratio=1.0, overlap="ipsep", mode="ipsep", splines="true")
            if agraph.number_of_nodes() <= 200:
                try:
                    agraph.layout(prog='neato', args="-Gepsilon=0.01 -Gmaxiter=50")
                except Exception as e:
                    raise RuntimeError("There was a problem with Graphviz. See https://graphviz.gitlab.io/") from e
            else:
                try:
                    agraph.layout(prog='dot')
                except Exception as e:
                    raise RuntimeError("There was a problem with Graphviz. See https://graphviz.gitlab.io/") from e
            if dest:
                svg_fn = dest + '.svg'
                png_fn = dest + '.png'
                agraph.draw(svg_fn,prog='neato',args="-Gepsilon=0.01 -Gmaxiter=50")
                agraph.draw(png_fn,prog='neato',args="-Gepsilon=0.01 -Gmaxiter=50")
            else:
                png_fn = self.filename[:-4] + "_graph.png"
                agraph.draw(png_fn,prog='neato',args="-Gepsilon=0.01 -Gmaxiter=50")
        else:
            raise RuntimeError("Nothing to draw.")


    def init_graph_to_file(self, dest=""):
        '''Saves image of graph generated by process step to file.

        Parameters
        ----------
        dest: str
            Destination for writing to file.
        '''
        return self._graph_to_file(self.init_graph,dest)

    def paths_graph_to_file(self, dest=""):
        '''Saves image of graph generated by pathways step to file.

        Parameters
        ----------
        dest:str
            Destination for writing to file.
        '''
        return self._graph_to_file(self.paths_graph,dest)

    def get_surface_exposed_residues(self):
        '''Returns list of surface exposed residues.

        Returns
        -------
        surface_exposed: list of str
            List of surface exposed residues identified by pyemap
        '''
        if self.init_graph:
            surface_exposed = []
            for n, d in self.init_graph.nodes(data=True):
                if d['shape'] == "box":
                    surface_exposed.append(n)
            return surface_exposed
        else:
            raise RuntimeError(
                "No graph found. Please run pyemap.process(my_emap) to generate the graph.")

    def report(self, dest=""):
        '''Returns report of most probable pathways. Writes to file if destination is specified.

        Parameters
        -----------
        dest: str, optional
            Destination for writing to file

        Returns
        --------
        output: str
            Formatted report of pathways found

        Raises
        -------
        RuntimeError
            Nothing to report
        '''
        if self.branches:
            output = ""
            for br in self.branches.values():
                output += str(br)
            if dest:
                fi = open(dest, "w")
                fi.write(output)
                return output
            else:
                return output
        else:
            raise RuntimeError("Nothing to report.")

    def _graph_to_Image(self, G):
        '''Returns PIL image of graph

        Returns
        --------
        img: :class:`PIL.Image.Image`
        '''
        if G:
            fout = tempfile.NamedTemporaryFile(suffix=".png")
            agraph = to_agraph(G)
            agraph.graph_attr.update(ratio=1.0, overlap="ipsep", mode="ipsep", splines="true")
            try:
                agraph.layout(prog='neato', args="-Gepsilon=0.01 -Gmaxiter=50")
            except Exception as e:
                raise RuntimeError("There was a problem with Graphviz. See https://graphviz.gitlab.io/") from e
            if agraph.number_of_nodes() <= 200:
                agraph.draw(fout.name, prog='neato')
            else:
                agraph.draw(fout.name, prog='dot')
            img = Image.open(fout.name)
            return img
        else:
            raise RuntimeError("Nothing to draw.")


    def init_graph_to_Image(self):
        '''Returns PIL image of initial graph

        Returns
        --------
        img: :class:`PIL.Image.Image`
        '''
        return self._graph_to_Image(self.init_graph)

    def paths_graph_to_Image(self):
        '''Returns PIL image of pathways graph
        
        Returns
        --------
        img: :class:`PIL.Image.Image`
        '''
        return self._graph_to_Image(self.paths_graph)




import pyemap
import os
import sys
from math import isclose
import unittest
from sys import platform

class PathwaysTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.my_emap = pyemap.parse(os.path.join(sys.path[0],"pyemap/tests/test_pdbs/2oal.pdb"))
        if platform == "linux":
            pyemap.process(cls.my_emap,sdef=1)
        elif platform == "darwin":
            pyemap.process(cls.my_emap)

    def test_yens_paths(self):
        pyemap.find_paths(self.my_emap,"Y167(B)",target="Y369(B)")
        assert len(self.my_emap.paths)==10
        pyemap.find_paths(self.my_emap,"Y167(B)",target="Y369(B)",max_paths=15)
        assert len(self.my_emap.paths)==15
        try:
            pyemap.find_paths(self.my_emap,"Y167(B)",target="Y167(A)")
            assert False
        except Exception as e:
            assert True

    def test_dijkstras_paths(self):
        pyemap.find_paths(self.my_emap,"Y167(B)")
        assert len(self.my_emap.paths)>0














# PyeMap: A python package for automatic identification of electron and hole transfer pathways in proteins.
# Copyright(C) 2017-2020 Ruslan Tazhigulov, James Gayvert, Ksenia Bravaya (Boston University, USA)
"""Processes parsed .pdb/.mmcif file, and generates a graph based on user selected options.

Collects requested residues and constructs a distance matrix, from which the graph is generated. Edges are filtered out based
on user specifications. Finally, the selected criteria for surface exposure is used to classify residues as buried or exposed.
Results are stored in the emap object which was passed in.

"""
import math
import sys
import Bio.PDB
import networkx as nx
import numpy as np
from Bio.PDB.DSSP import DSSP
from Bio.PDB.ResidueDepth import get_surface, residue_depth
from scipy.spatial import distance_matrix
import warnings
from .data import res_name_to_char, TRP_sc, TYR_sc, PHE_sc, HIS_sc

# Monkey patches detach self to save original ID upon re-assignment to custom residue


def detach_parent(self):
    if self.parent:
        self.original_id = self.parent.full_id
    self.parent = None


Bio.PDB.Atom.Atom.detach_parent = detach_parent

# Monkey patches Atom __init__ to include a variable which tracks its original ID and assigns a unique ID


def atm_init(self, name, coord, bfactor, occupancy, altloc, fullname, serial_number, element=None):
    self.level = "A"
    # Reference to the residue
    self.parent = None
    # the atomic data
    self.name = name  # eg. CA, spaces are removed from atom name
    self.fullname = fullname  # e.g. " CA ", spaces included
    self.coord = coord
    self.bfactor = bfactor
    self.occupancy = occupancy
    self.altloc = altloc
    # (structure id, model id, chain id, residue id, atom id)
    self.full_id = None
    # id of atom is the atom name (e.g. "CA")
    self.id = name + str(serial_number)
    self.disordered_flag = 0
    self.anisou_array = None
    self.original_id = None
    self.siguij_array = None
    self.sigatm_array = None
    self.serial_number = serial_number
    # Dictionary that keeps additional properties
    self.xtra = {}
    assert not element or element == element.upper(), element
    self.element = self._assign_element(element)
    self.mass = self._assign_atom_mass()
    # For atom sorting (protein backbone atoms first)
    self._sorting_keys = {'N': 0, 'CA': 1, 'C': 2, 'O': 3}


Bio.PDB.Atom.Atom.__init__ = atm_init

# monkey patches mangling of disordered atoms


def get_unpacked_list(self):
    """
     Returns all atoms from the residue,
     in case of disordered, keep only first alt loc and remove the alt-loc tag
    """
    atom_list = self.get_list()
    undisordered_atom_list = []
    for atom in atom_list:
        if atom.is_disordered():
            atom.altloc = " "
            undisordered_atom_list.append(atom)
        else:
            undisordered_atom_list.append(atom)
    return undisordered_atom_list


Bio.PDB.Residue.Residue.get_unpacked_list = get_unpacked_list


def pathways_model(dist, coef_alpha, exp_beta, r_offset):
    """Applies penalty function parameters and returns score.

        :math:`\\epsilon =\\alpha \\exp(-\\beta(R-R_{offset}))`

    Parameters
    ----------
    dist: float
        Actual distance in angstroms
    coef_alpha,exp_beta,r_offset:float
        Penalty function parameters

    Returns
    -------
    mod_penalty: float
    
    """
    penalty = coef_alpha * np.exp(-exp_beta * (dist - r_offset))
    mod_penalty = -np.log10(penalty)
    return mod_penalty


def calculate_residue_depth(aromatic_residues, model):
    """Returns a list of surface exposed residues as determined by residue depth.

    Parameters
    ----------
    aromatic_residues: list of :class:`Bio.PDB.Residue.Residue`
        residues included in the analysis
    model: :class:`Bio.PDB.Model.Model`
        BioPython object containing a model of a PDB or MMCIF file

    Returns
    -------
    surface_exposed_res: list of str
        List of residue names corresponding to the surface exposed residues

    """
    try:
        surface = get_surface(model)
        cutoff = 3.03
        surface_exposed_res = []
        for residue in aromatic_residues:
            depth = residue_depth(residue, surface)
            if (depth <= cutoff):
                surface_exposed_res.append(residue.node_label)
        return surface_exposed_res
    except Exception as e:
        warnings.warn("Unable to calculate residue depth. Check that MSMS is installed. Please note that MSMS is not compatible with MacOS Catalina.", RuntimeWarning,stacklevel=2)
        return []


def calculate_asa(model, filename, node_list):
    """Returns a list of surface exposed residues as determined by relative solvent accessibility.

    Only standard protein residues are currently supported. Non-protein and user specified custom residues cannot be
    classified as surface exposed using this criteria.

    Parameters
    ---------
    model: :class:`Bio.PDB.Model.Model`
        Model which contains chains and residues of protein strucutre
    filename: str
        Name of pdb file to be analyzed
    AROM_LIST : list of str
        List containing which standard residues are included in analysis
    chain_list: list of str
        Chains are included in analysis

    Notes
    -----
    The relative accessible surface area (RSA) of each residue is calculated using the Bio.PDB.DSSP module. A residue
    with an RSA value of 0.05 or higher is classified as surface exposed.

    References
    ---------
    Tien, M. Z.; Meyer, A. G.; Sydykova, D. K.; Spielman, S. J.; Wilke, C. O. PLoS ONE 2013, 8 (11).
        Reference for relative solvent accessibility cutoff of 0.05, and for MaxASA values

    """
    cutoff = .05
    surface_exposed_res = []
    try:
        dssp = DSSP(model, filename, acc_array="Wilke")
        keys = list(dssp.keys())
        for key in keys:
            goal_str = dssp[key][1] + str(key[1][1]) + "(" + str(key[0]) + ")"
            if goal_str in node_list and dssp[key][3] >= cutoff:
                surface_exposed_res.append(goal_str)
    except Exception as e:
        warnings.warn("Unable to calculate solvent accessibility. Check that DSSP is installed.", RuntimeWarning,stacklevel=2)
    return surface_exposed_res


def dist(x, y):
    """Returns Euclidean distance between two 3 dimensional points.

    Parameters
    ----------
    x: 3D numpy array of float
        Point 1
    y: 3D numpy array of float
        Point 2

    Returns
    -------
    float:
        Euclidean distance between two points

    """
    return np.sqrt(np.sum((x - y)**2))

def get_atom_list(res):
    """ Recovers side chain atoms only of standard aromatic residues, returns all atoms for other residues.
    
    Parameters
    -----------
    res: :class:`Bio.PDB.Residue.Residue`
        A BioPython residue object
    
    Returns
    -------
    atom_list: list of :class:`Bio.PDB.Atom.Atom`
        List of atoms to be used in distance matrix calculation
    """
    atom_list = []
    for atm in res.get_atoms():
        if res.resname == "TRP" and atm.name in TRP_sc:
            atom_list.append(atm)
        elif res.resname == "TYR" and atm.name in TYR_sc:
            atom_list.append(atm)
        elif res.resname == "PHE" and atm.name in PHE_sc:
            atom_list.append(atm)
        elif res.resname == "HIS" and atm.name in HIS_sc:
            atom_list.append(atm)
        elif res.resname not in ["TRP","TYR","PHE","HIS"]:
            atom_list.append(atm)
    return atom_list

def get_full_atom_distance_matrix(residues):
    com_d = []
    atoms_per_res = []
    for i in range(len(residues)):
        res = residues[i]
        res.get_full_id()
        atm_list = get_atom_list(res)
        atoms_per_res.append(len(atm_list))
        for j in range(len(atm_list)):
            cur_atom = atm_list[j]
            com_d.append(np.array([cur_atom.coord[0], cur_atom.coord[1], cur_atom.coord[2]]))
    return distance_matrix(com_d, com_d), atoms_per_res


def get_com_distance_matrix(residues):
    com_d = []
    # Compute COMs (x0, y0, z0) of side-chains
    for i in range(len(residues)):
        res = residues[i]
        res.get_full_id()
        atm_list = get_atom_list(res)
        x_wsum, y_wsum, z_wsum, mass_sum = 0.0, 0.0, 0.0, 0.0
        for j in range(len(atm_list)):
            cur_atom = atm_list[j]
            mass = cur_atom.mass
            x = cur_atom.coord[0]
            y = cur_atom.coord[1]
            z = cur_atom.coord[2]
            x_wsum += x * mass
            y_wsum += y * mass
            z_wsum += z * mass
            mass_sum += mass
        com_x = x_wsum / mass_sum
        com_y = y_wsum / mass_sum
        com_z = z_wsum / mass_sum
        com_d.append(np.array([com_x, com_y, com_z]))
    return distance_matrix(com_d, com_d)


def closest_atom_dmatrix(residues, coef_alpha, exp_beta, r_offset):
    """Constructs distance matrix based on closest atom distance.
    
    Parameters
    ----------
    residues: list of :class:`Bio.PDB.Residue.Residue`
        List of BioPython residues
    coef_alpha,exp_beta,r_offset:float
        Penalty funciton parameters
    Returns
    -------
    distance_matrix: numpy.array of float
        Distance matrix of residues
    pathways_matrix: numpy.array of float
        Modified distance matrix of residues using scores determined by penalty function parameters
    """
    dmat, atoms_per_res = get_full_atom_distance_matrix(residues)
    distance_matrix = np.zeros((len(residues), len(residues)))
    pathways_matrix = np.zeros((len(residues), len(residues)))
    slice1_idx = 0
    for i in range(0, len(residues)):
        slice2_idx = slice1_idx + atoms_per_res[i]
        slice3_idx = slice2_idx
        for j in range(i + 1, len(residues)):
            slice4_idx = slice3_idx + atoms_per_res[j]
            my_slice = dmat[slice1_idx:slice2_idx, slice3_idx:slice4_idx]
            min_val = np.min(my_slice)
            distance_matrix[i][j] = min_val
            distance_matrix[j][i] = min_val
            pathways_matrix[i][j] = pathways_model(min_val, coef_alpha, exp_beta, r_offset)
            pathways_matrix[j][i] = pathways_matrix[i][j]
            slice3_idx = slice4_idx
        slice1_idx = slice2_idx
    return distance_matrix, pathways_matrix


def com_dmatrix(residues, coef_alpha, exp_beta, r_offset):
    """Constructs distance matrix based on distances between centers of mass.

    Parameters
    ----------
    residues: list of :class:`Bio.PDB.Residue.Residue`
        List of residues to be included in analysis
    coef_alpha,exp_beta,r_offset:float
        Penalty funciton parameters

    Returns
    -------
    node_label: dict of int:str
        List of node labels for graph
    distance_matrix: numpy array of :class:`Bio.PDB.Residue.Residue`
        Distance matrix of residues

    """
    com_d = get_com_distance_matrix(residues)
    # calculate matrix with penalty functions
    pathways_matrix = np.zeros((len(com_d), len(com_d)))
    for i in range(len(com_d)):
        for j in range(i + 1, len(com_d)):
            dist_i_j = com_d[i][j]
            pathways_matrix[i][j] = pathways_model(dist_i_j, coef_alpha, exp_beta, r_offset)
            pathways_matrix[j][i] = pathways_matrix[i][j]
    return com_d, pathways_matrix


def process_standard_residues(standard_residue_list):
    """Generates customized Bio.PDB.Residue.Residue objects. Residues containing no side chain atoms will be removed.

    Parameters
    ----------
    standard_residue_list: list of :class:`Bio.PDB.Residue.Residue`
        List of Bio.PDB.Residue.Residue objects corresponding to protein residues

    Returns
    -------
    res_list: list of :class:`Bio.PDB.Residue.Residue`

    """
    res_list = []
    for i in range(0, len(standard_residue_list)):
        res = standard_residue_list[i]
        res_letter = res_name_to_char.get(res.resname)
        chain = res.full_id[2]
        resnum = res.full_id[3][1]
        res.node_label = res_letter + str(resnum) + "(" + chain + ")"
        atm_ids = []
        atm_names = []
        valid = False
        for atm in res.get_atoms():
            atm_ids.append(atm.id)
            atm_names.append(atm.name)
        for k in range(0, len(atm_ids)):
            if res.resname == "TRP":
                if atm_names[k] in TRP_sc:
                    valid = True
            elif res.resname == "TYR":
                if atm_names[k] in TYR_sc:
                    valid = True
                    break
            elif res.resname == "PHE":
                if atm_names[k] in PHE_sc:
                    valid = True
                    break
            elif res.resname == "HIS":
                if atm_names[k] in HIS_sc:
                    valid = True
                    break
        if valid:
            res_list.append(res)
        else:
            if res.resname == "TRP":
                sca = str(TRP_sc)
            elif res.resname == "TYR":
                sca = str(TYR_sc)
            elif res.resname == "PHE":
                sca = str(PHE_sc)
            elif res.resname == "HIS":
                sca = str(HIS_sc)
            side_chain_atms = sca
            def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
                        return '%s:%s: %s: %s\n' % (filename, lineno, category.__name__, message)
            warnings.formatwarning = warning_on_one_line
            warnings.warn("The record for residue " + res.node_label + " did not contain any of the following side chain atoms: " + side_chain_atms + " and therefore is not included in the graph.",RuntimeWarning,stacklevel=2)
    return res_list


def create_user_res(serial_list, all_atoms, chain_selected, used_atoms, user_res_names):
    """Creates a customized BioPython Residue object corresponding to a user specified residue.

    Parameters
    ----------
    serial_list: list of int
        List of atom serial numbers included in residue
    all_atoms: list of :class:`Bio.PDB.Atom.Atom`
        All atoms in protein on selected chains
    chain_selected: list of str
        Chains included in analysis
    used_atoms: list of :class:`Bio.PDB.Atom.Atom`
        Atoms already included in analysis
    user_res_names: list of str
        User residue names already included in the analysis.

    Returns
    -------
    user_res: :class:`Bio.PDB.Residue.Residue`
        Customized residue object corresponding to the atoms in serial_list

    """
    source_res = []
    atm_list = []
    for atm in all_atoms:
        if atm.serial_number in serial_list:
            if atm.serial_number in used_atoms:
                message = "Invalid atom serial number range. Atom " + str(
                    atm.serial_number) + " is already included in another residue."
                raise ValueError(message)
            if atm.parent.parent.id not in chain_selected:
                message = "Invalid atom serial number range. Atom " + str(
                    atm.serial_number) + " is in chain " + str(atm.parent.parent.id) + "."
                raise ValueError(message)
            atm.get_full_id()
            if not source_res:
                source_res = atm.parent
            atm.get_full_id()
            atm_copy = atm.copy()
            atm_list.append(atm_copy)
    source_res.get_full_id()
    user_res = source_res.copy()
    for atm in list(source_res.get_atoms()):
        user_res.detach_child(atm.id)
    for atm in atm_list:
        user_res.add(atm)
    k = 1
    name = "CUST"
    name += "-"
    while name + str(k) in user_res_names:
        k += 1
    user_res_names.append(name + str(k))
    user_res.resname = name + str(k)
    user_res.node_label = user_res.resname
    return user_res


def get_standard_residues(all_residues, chain_list, include_Trp, include_Tyr, include_Phe, include_His):
    """Returns list of standard aromatic residues.

    Runs through every residue in the structure, keeping only the TRP, TYR
    (optionally PHE and HIS if specified by user) residues on the chains
    that were chosen by the user.

    Parameters
    ----------
    all_residues: list of :class:`Bio.PDB.Residue.Residue`
        List of every residue in structure
    chain_list: list of str
        Chains to be included in analysis
    include_Trp,include_Tyr,include_Phe,include_His: bool
        True if residue type is included

    Returns
    -------
    residue_list: list of :class:`Bio.PDB.Residue.Residue`
        Residues to be included in analysis
    AROM_LIST: list of str
        Types of aromatic residues included in graph

    """
    AROM_LIST = ['TRP', 'HIS', 'PHE', 'TYR']
    if not include_Trp:
        AROM_LIST.remove('TRP')
    if not include_Tyr:
        AROM_LIST.remove('TYR')
    if not include_Phe:
        AROM_LIST.remove('PHE')
    if not include_His:
        AROM_LIST.remove('HIS')
    residue_list = []
    for res in all_residues:
        if res.resname in AROM_LIST and res.parent.id in chain_list:
            res.get_full_id()
            arom_res = res.copy()
            arom_res.get_full_id()
            residue_list.append(arom_res)
    residue_list = process_standard_residues(residue_list)
    return residue_list, AROM_LIST


def get_user_residues(custom, all_atoms, chain_selected, used_atoms):
    """Generates customized Bio.PDB.Residue.Residue objects from a custom atom string.

    Users may not choose atoms that are already part of standard residues or eta moieties, nor those that are not
    part of the selected chains.

    Parameters
    ----------
    custom: str
        Specified by user to select atoms for custom residues
    all_atoms: list of :class:`Bio.PDB.Atom.Atom`
        All atoms in protein on selected chains
    chain_selected: list of str
        Chains included in analysis
    used_atoms: list of :class:`Bio.PDB.Atom.Atom`
        Atoms already included in analysis

    Returns
    -------
    res_list: list of :class:`Bio.PDB.Residue.Residue`
        Custom residues specified by user

    Notes
    -----
    The custom atom string syntax is as follows:
    Custom residues are specified based on PDB atom serial number. Ranges are specified with a dash (ex. 10-20).
    Individual atoms may also be selected, separated by commas (ex. 1-10,15,17). To select multiple custom residues,
    the user encloses the atom ranges with parentheses, and then separate them with commas. In the graph and
    visualization, these residues are named CUS-1, CUS-2 etc. based on the order they were specified.

    """
    user_res_names = []
    custom = custom.strip()
    if custom:
        res_list = []
        l1 = custom.split("),(")
        for a in l1:
            serial_number_list = []
            # get rid of the (
            atm_str = a[:]
            atm_str = atm_str.replace("(", '')
            atm_str = atm_str.replace(")", '')
            l2 = atm_str.split(",")
            for atm in l2:
                if "-" in atm:
                    index = atm.index("-")
                    start_atm = int(atm[:index])
                    end_atm = atm[index + 1:]
                    end_atm = end_atm.replace(")", "")
                    end_atm = int(end_atm)
                    for i in range(start_atm, end_atm + 1):
                        serial_number_list.append(i)
                else:
                    serial_number_list.append(int(atm))
            if serial_number_list:
                new_res = create_user_res(
                    serial_number_list, all_atoms, chain_selected, used_atoms, user_res_names)
            else:
                raise SyntaxError(
                    "Invalid atom serial number range. See the manual for proper syntax.")
            res_list.append(new_res)
            for atm in new_res.get_atoms():
                used_atoms.append(atm.serial_number)
        return res_list
    return []


def finish_graph(G, surface_exposed_res, chain_list):
    """Sets surface exposed residues as boxes in graph and removes disconnected vertices.

    Parameters
    ----------
    G: :class:`networkx.graph`
        Graph object constructed from distance matrix of residues
    surface_exposed_res: list of str
        Names of surface exposed residues
    chain_list: list of str
        Names of chains included in analysis

    """
    for goal in surface_exposed_res:
        G.nodes[goal]['margin'] = '0.11'
        G.nodes[goal]['shape'] = 'box'
    # get rid of all disconnected nodes
    all_nodes = list(G.nodes())
    for node in all_nodes:
        if G[node] == {}:
            G.remove_node(node)


def filter_edges(G, G_pathways, distance_cutoff, percent_edges, num_st_dev_edges):
    '''Applies distance based filters to edges, removing those edges which do not fit criteria.

    Parameters
    ----------
    G: :class:`networkx.graph`
        graph where edge weights are pure distances
    G_pathways: :class:`networkx.graph`
        graph where edge weights are distance dependent penalty functions
    distance_cutoff,percent_edges,num_st_dev_edges: float
        Parameters that determine which edges are kept.
    '''
    # keep 99th percentile and ~20A filter on all edges
    included_edges = []
    for node in G.nodes():
        edge_length_per_node = []
        weights = []
        for neighbor in G[node]:
            weights.append(G.get_edge_data(node, neighbor)['weight'])
        weights = sorted(weights)
        thresh_index = math.ceil(len(weights) * percent_edges / 100)
        for neighbor in G[node]:
            if weights.index(G.get_edge_data(node, neighbor)['weight']) <= thresh_index and \
                    G.get_edge_data(node, neighbor)['weight'] <= distance_cutoff:
                edge_length_per_node.append(
                    G.get_edge_data(node, neighbor)['weight'])

        len_average, len_st_dev = 0.0, 0.0
        if edge_length_per_node != []:
            edge_length_per_node = np.array(
                edge_length_per_node, dtype='float64')
            len_average = np.average(edge_length_per_node)
            len_st_dev = np.std(edge_length_per_node)

        for neighbor in G[node]:
            if weights.index(G.get_edge_data(node, neighbor)['weight']) <= thresh_index and \
                    G.get_edge_data(node, neighbor)['weight'] <= distance_cutoff and \
                    np.round(G.get_edge_data(node, neighbor)['weight'], 8) <= np.round((len_average + num_st_dev_edges * len_st_dev), 8):
                included_edges.append([node, neighbor])

    excluded_edges = []
    for edge in G.edges():
        node1 = edge[0]
        node2 = edge[1]
        if ([node1, node2] not in included_edges) and ([node2, node1] not in included_edges):
            excluded_edges.append(edge)

    for edge in excluded_edges:
        node1 = edge[0]
        node2 = edge[1]
        if G.has_edge(node1, node2):
            G.remove_edge(node1, node2)
        if G_pathways.has_edge(node1, node2):
            G_pathways.remove_edge(node1, node2)


def create_graph(dmatrix, pathways_matrix, node_labels, distance_cutoff, percent_edges, num_st_dev_edges, eta_moieties):
    """Constructs the graph from the distance matrix and node labels.

    Parameters
    ----------
    dmatrix: numpy.array of float
        Distance matrix of aromatic residues.
    node_label: list of str
        Labels for residues in the graph.
    distance_cutoff,percent_edges,num_st_dev_edges: float
        Parameters that determine which edges are kept.
    eta_moieties: list of str
        Non standard residues that were automatically identified

    Returns
    -------
    G: :class:`networkx.Graph`
        Graph of aromatic residues in protein

    References
    ----------
    Gray, H. B.; Winkler, J. R. Long-Range Electron Transfer. Proc. Natl. Acad. Sci. U. S. A. 2005, 102 (10),
        3534 LP-3539.
        Reference for 20A filter on edges
    """
    np.set_printoptions(threshold=sys.maxsize)
    G = nx.from_numpy_array(dmatrix)
    minval_pathways = np.min(pathways_matrix[pathways_matrix.nonzero()])
    G_pathways = nx.from_numpy_array(pathways_matrix)
    filter_edges(G, G_pathways, distance_cutoff,
                 percent_edges, num_st_dev_edges)
    for u, v, d in G_pathways.edges(data=True):
        d['len'] = d['weight'] / minval_pathways
        d['distance'] = G[u][v]['weight']
    G = G_pathways
    G = nx.relabel_nodes(G, node_labels)
    for name_node in G.nodes():
        G.nodes[name_node]['style'] = 'filled'
        G.nodes[name_node]['fontname'] = 'Helvetica-Bold'
        G.nodes[name_node]['fontsize'] = 32
        G.nodes[name_node]['shape'] = "oval"
        G.nodes[name_node]['margin'] = '0.04'
        G.nodes[name_node]['fontcolor'] = "#000000"
        G.nodes[name_node]['color'] = '#708090'
        G.nodes[name_node]['penwidth'] = 2.0
        if(name_node[1].isdigit()):
            if name_node not in eta_moieties:
                if 'Y' == name_node[0]:
                    G.nodes[name_node]['fillcolor'] = '#96c8f0'
                elif 'W' == name_node[0]:
                    G.nodes[name_node]['fillcolor'] = '#f07878'
                elif 'F' == name_node[0]:
                    G.nodes[name_node]['fillcolor'] = '#f09664'
                elif 'H' == name_node[0]:
                    G.nodes[name_node]['fillcolor'] = '#c8f0c8'
            else:
                G.nodes[name_node]['fillcolor'] = '#FFC0CB'
        else:
            G.nodes[name_node]['fillcolor'] = '#FFC0CB'
    for edge in G.edges():
        name_node1, name_node2 = edge[0], edge[1]
        G[name_node1][name_node2]['color'] = '#778899'
        G[name_node1][name_node2]['penwidth'] = 1.5
        G[name_node1][name_node2]['style'] = 'dashed'
    return G


def process(emap,
            chains="All",
            eta_moieties="All",
            dist_def=0,
            sdef=1,
            trp=True,
            tyr=True,
            phe=False,
            his=False,
            custom="",
            distance_cutoff=20,
            percent_edges=1.0,
            num_st_dev_edges=1.0,
            coef_alpha=1.0,
            exp_beta=2.3,
            r_offset=0.0):
    """Constructs emap graph theory model based on user specs, and saves it to the emap object.

    Parameters
    ---------
    emap: :class:`~pyemap.emap`
        Object for storing state of emap analysis.
    chains: list of str
        List of strings corresponding to chains included in analysis
    eta_moieties: list of str
        List of strings corresponding to residue names of eta moieties
    dist_def: int, optional
        0 for center of mass, 1 for closest atom
    sdef: int, optional
        0 for residue depth, 1 for solvent accessibility
    trp,tyr,phe,his: bool, optional
        True if residue type is included
    custom: str, optional
        Custom atom string specified by user
    distance_cutoff: float
         Defines a pure distance threshold. PyeMap will only keep edges with distances less than or equal distance_cutoff.
    percent_edges: float
        Specifies a percentage of the shortest edges per vertex to keep.
    num_st_dev_edges: float
        For a particular vertex, only edges with length < average length + num_st_dev_edges * SD included
    coef_alpha,exp_beta,r_offset: float, optional
        Penalty function parameters.
    Raises
    ------
    RuntimeError:
        Not enough residues to construct a graph

    """
    emap._reset_process()
    if eta_moieties == "All":
        eta_moieties = emap.eta_moieties.keys()
    if chains == "All":
        chains = emap.chains
    model = emap.structure[0]
    all_residues = list(model.get_residues())
    all_atoms = list(model.get_atoms())
    aromatic_residues, AROM_LIST = get_standard_residues(
        all_residues, chains, trp, tyr, phe, his)
    for resname in eta_moieties:
        AROM_LIST.append(resname)
        aromatic_residues.append(emap.eta_moieties[resname])
    used_atoms = []
    for res in aromatic_residues:
        for atm in res.get_atoms():
            used_atoms.append(atm.serial_number)
    user_residues = []
    if custom:
        user_residues = get_user_residues(
            custom, all_atoms, chains, used_atoms)
    for res in user_residues:
        AROM_LIST.append(res.resname)
        emap.user_residues[res.resname] = res
    aromatic_residues += user_residues
    node_labels = {}
    residue_numbers = {}
    aligned_residue_numbers = {}
    for i in range(0, len(aromatic_residues)):
        node_labels[i] = aromatic_residues[i].node_label
        residue_numbers[aromatic_residues[i].node_label] = aromatic_residues[i].full_id[3][1]
        if hasattr(aromatic_residues[i],'aligned_residue_number'):
            aligned_residue_numbers[aromatic_residues[i].node_label] = aromatic_residues[i].aligned_residue_number
    if int(dist_def) == 0:
        dmatrix, pathways_matrix = com_dmatrix(aromatic_residues, coef_alpha, exp_beta, r_offset)
    else:
        dmatrix, pathways_matrix = closest_atom_dmatrix(aromatic_residues, coef_alpha, exp_beta, r_offset)
    G = create_graph(dmatrix, pathways_matrix, node_labels,
                     distance_cutoff, percent_edges, num_st_dev_edges,emap.eta_moieties.keys())
    if len(G.edges()) == 0:
        raise RuntimeError(
            "Not enough edges to construct a graph.")
    # define surface exposed residues
    if int(sdef) == 0:
        surface_exposed_res = calculate_residue_depth(aromatic_residues, model)
    else:
        pdb_file = emap.filename
        surface_exposed_res = calculate_asa(model, pdb_file, node_labels.values())
    finish_graph(G, surface_exposed_res, chains)
    for res in aromatic_residues:
        emap._add_residue(res)
    emap._store_initial_graph(G)



# PyeMap: A python package for automatic identification of electron and hole transfer pathways in proteins.
# Copyright(C) 2017-2020 Ruslan Tazhigulov, James Gayvert, Ksenia Bravaya (Boston University, USA)
"""Finds shortest paths in graph given a source and optionally a target node.

Defines implementations of yen's and dijkstra's algorithms for calculating the shortest path(s) from
the source to target/surface exposed residues. Also defines ShortestPath and Branch objects for
organizing the pathways based on their distances and first surface exposed residue reached during
the pathway.

"""
import itertools
import string
import networkx as nx
from functools import total_ordering

@total_ordering
class ShortestPath(object):
    """Data structure used to store shortest paths.

    Contains functions for comparison and string representation, both for the user output and
    in NGL viewer selection language for the purpose of visualization. Sorting for ShortestPath
    objects is done based on the length attribute.

    Attributes
    ----------
    path: list of str
        List of residue names that make up the shortest path
    path_id: list of str
        List of residues that make up the shortest path
    length: float
        Total distance from source to target
    selection_strs: list of str
        NGL selection strings for visualization
    color_list: list of str
        Colors of residues in visualization
    labeled_atoms: list of str
        Atom names which are labeled in NGL visualization
    label_texts: list of str
        Labels of residues in NGL visualization
    """

    def __init__(self, path, length):
        '''Initializes ShortestPath object.

        Parameters
        ----------
        path: list of str
            List of residues that make up the shortest path
        length: float
            Total distance from source to target
        '''
        self.path = path
        self.length = length
        self.path_id = "none"
        self.selection_strs = []
        self.color_list = []
        self.labeled_atoms = []
        self.label_texts = []

    def __eq__(self, other):
        return self.length == other.length

    def __lt__(self, other):
        return self.length < other.length

    def __str__(self):
        printline = self.path_id + ": " + \
            str(self.path) + " " + str('{:.2f}'.format(round(self.length, 2)))
        return printline

    def get_path_as_list(self):
        original_list = [[self.path_id], self.path, [
            str('{:.2f}'.format(round(self.length, 2)))]]
        merged = list(itertools.chain(*original_list))
        return merged

    def set_id(self, path_id):
        """Setter for path_id"""
        self.path_id = path_id

    def set_visualization(self, selection_strs, color_list, labeled_atoms, label_texts):
        '''Saves information needed for NGL visualization.'''
        self.selection_strs = selection_strs
        self.color_list = color_list
        self.labeled_atoms = labeled_atoms
        self.label_texts = label_texts


class Branch(object):
    """Data structure used to group shortest paths with a common first surface exposed residue.

    Shortest paths are classified into branches based on the first surface exposed residue
    reached during the course of the pathway. For example, given a source node A, and
    surface exposed nodes B and C, the paths [A,F,B] and [A,F,B,C] will both be part of the
    "B" branch. The path [A,E,C] would be part of its own 'C" branch.

    Attributes
    ----------
    branch_id: str
        Unique identifier for a branch
    target: str
        Target node which a branch corresponds to
    paths: list of :class:`~pyemap.ShortestPath`
        List of ShortestPath objects that make up a branch
    """

    def __init__(self, branch_id, target):
        self.branch_id = branch_id
        self.target = target
        self.paths = []

    def add_path(self, path):
        """Adds a path to the branch and sets the path_id.

        Each time a path is added, the paths in the branch are sorted. After sorting, each path is assigned a
        path id composed of the branch id and its location in the paths list. For example, the shortest path
        in branch 12 would be assigned the id '12a', the second shortest '12b' and so on.

        Parameters
        ----------
        path: :class:`~pyemap.ShortestPath`
            A ShortestPath from source to a surface exposed residue
        """
        letters = list(string.ascii_letters)
        if len(self.paths) < len(letters):
            self.paths.append(path)
            self.paths = sorted(self.paths)
            i = 0
            while i < len(self.paths) and i < len(letters):
                self.paths[i].set_id(str(self.branch_id) + str(letters[i]))
                i += 1

    def __str__(self):
        ''' String representation of Branch. First line is: "Branch: `branch_id`" and subsequent lines
        are the string representations of each ShortestPath object comprising the Branch.
        '''
        printline = "Branch: " + str(self.target)
        for pt in self.paths:
            printline += "\n" + str(pt)
        printline += "\n"
        return printline

    def get_branch_as_list(self):
        ''' List representation of Branch. First entry is: "Branch: `branch_id`" and the rest of the entries
        are the string representations of each ShortestPath object comprising the Branch.
        '''
        branch_list = []
        printline = "Branch: " + str(self.target)
        branch_list.append([printline])
        for pt in self.paths:
            branch_list.append(pt.get_path_as_list())
        return branch_list


def _is_parent_pathway(shortest_path, targets):
    """Returns true if ShortestPath is a parent pathway, false if not.

    A ShortestPath object is the parent of a branch if its terminal residue is the
    only surface exposed residue in the path. For example, if targets=[A,B,C] and
    the pathway is [H,I,C], then this pathway is a parent pathway. In contrast, if
    the pathway is [H,B,A], then this pathway is not a parent pathway.

    Parameters
    ----------
    shortest_path: ShortestPath
        ShortestPath object
    targets: list of str
        List of surface exposed residues

    Returns
    -------
    bool
        True if path is ShortestPath object is a parent pathway
    """

    count = 0
    for res in shortest_path.path:
        if res in targets:
            count += 1
    return count == 1


def _find_branch(pt, targets, branches):
    """Determines which branch a pathway belongs to and returns that branch.

    A ShortestPath belongs to a branch if the first surface exposed residue it reaches during the
    pathway is the target of that particular branch.

    Parameters
    ----------
    pt: ShortestPath
        ShortestPath object
    targets: list of str
        List of surface exposed residues
    branches: list of pyemapBranch objects
        List of branches already found

    Returns
    -------
    cur_branch: Branch
        Branch object that pt belongs to

    """
    res = pt.path[0]
    count = 0
    while res not in targets:
        count += 1
        res = pt.path[count]
    count = 0
    cur_branch = branches[0]
    while not res == cur_branch.target:
        count += 1
        cur_branch = branches[count]
    return cur_branch


def dijkstras_shortest_paths(G, start, targets):
    """Returns shortest path from source to each surface exposed residue.

    Performs Dijkstra's algorithm from the source to each surface exposed residue, finding the
    shortest path. The ShortestPath objects are organized into branches based on the first surface
    exposed residue reached during the course of the pathway.

    Parameters
    ----------
    G: :class:`networkx.Graph`
        Undirected, weighted residue graph
    start: str
        Source node
    targets: list of str
        List of surface exposed residues

    Returns
    -------
    branches: list of :class:`~pyemap.Branch`
        A list of Branch objects representing the groups of pathways found

    Raises
    ------
    RuntimeError:
        No shortest paths to surface found

    """
    shortestPaths = []
    for goal in targets:
        path = []
        try:
            path = nx.dijkstra_path(G, start, goal)
        except Exception as e:
            path = []
        if not path == []:
            sum = 0
            for i in range(0, len(path) - 1):  # sum up edge weights
                sum += (G[path[i]][path[i + 1]]['weight'])
            shortestPaths.append(ShortestPath(path, sum))
    shortestPaths = sorted(shortestPaths)
    branches = []
    # find the parent pathways
    for pt in shortestPaths:
        if _is_parent_pathway(pt, targets):
            path = pt.path
            for i in range(0, len(path) - 1):
                G[path[i]][path[i + 1]]['color'] = '#778899FF'
                G[path[i]][path[i + 1]]['penwidth'] = 6.0
                G[path[i]][path[i + 1]]['style'] = 'solid'
                G.nodes[path[i]]['penwidth'] = 6.0
                G.nodes[path[i + 1]]['penwidth'] = 6.0
                # make the nodes look opaque if they are connected to the source
                if len(G.nodes[path[i]]['fillcolor']) != 9:
                    G.nodes[path[i]]['fillcolor'] += 'FF'
                    G.nodes[path[i]]['color'] = '#708090FF'
                if len(G.nodes[path[i + 1]]['fillcolor']) != 9:
                    G.nodes[path[i + 1]]['fillcolor'] += 'FF'
                    G.nodes[path[i + 1]]['color'] = '#708090FF'
            br = Branch(len(branches) + 1, pt.path[-1])
            branches.append(br)
            br.add_path(pt)
    # find the sub pathways
    for pt in shortestPaths:
        if not _is_parent_pathway(pt, targets):
            _find_branch(pt, targets, branches).add_path(pt)
            path = pt.path
            for i in range(0, len(path) - 1):
                if G[path[i]][path[i + 1]]['color'] != '#778899FF':
                    G[path[i]][path[i + 1]]['color'] = '#7788995F'
                G[path[i]][path[i + 1]]['penwidth'] = 6.0
                G[path[i]][path[i + 1]]['style'] = 'solid'
                G.nodes[path[i]]['penwidth'] = 6.0
                G.nodes[path[i + 1]]['penwidth'] = 6.0
                # make the nodes look opaque if they are connected to the source
                if len(G.nodes[path[i]]['fillcolor']) != 9:
                    G.nodes[path[i]]['fillcolor'] += '5F'
                    G.nodes[path[i]]['color'] = '#7080905F'
                if len(G.nodes[path[i + 1]]['fillcolor']) != 9:
                    G.nodes[path[i + 1]]['fillcolor'] += '5F'
                    G.nodes[path[i + 1]]['color'] = '#7080905F'
    if len(shortestPaths) == 0:
        raise RuntimeError("No paths to the surface found.")
    return branches


def yens_shortest_paths(G, start, target, max_paths=10):
    """Returns top 5 shortest paths from source to target.

    Uses Yen's algorithm to calculate the shortest paths from source to target, writes
    out the ShortestPath objects to file, and returns the 5 pathway IDs. In the graph, nodes and
    edges that are part of any pathways are made opaque, and the shortest path is highlighted.


    Parameters
    ----------
    G: :class:`networkx.Graph` object
        Undirected, weighted residue graph
    start: str
        Source node
    target: str
        Target node
    max_paths: int, optional
        Maximum number of paths to search for

    Returns
    -------
    list of :class:`~pyemap.Branch` objects
        A list of length 1 containing a single Branch object which represents the group of pathways found.

    References
    ----------
    Jin Y. Yen, Finding the K Shortest Loopless Paths in a Network, Management Science,
    Vol. 17, No. 11, Theory Series (Jul., 1971), pp. 712-716.

    Raises
    ------
    RuntimeError:
        No shortest paths to target found.

    """
    letters = list(string.ascii_letters)
    shortestPaths = []
    k = 0
    paths = list(itertools.islice(nx.shortest_simple_paths(G, start, target), max_paths))
    for k in range(0, len(paths)):
        path = paths[k]
        sum = 0
        for i in range(0, len(path) - 1):  # sum up edge weights
            sum += (G[path[i]][path[i + 1]]['weight'])
        path = ShortestPath(path, sum)
        shortestPaths.append(path)
    if shortestPaths:
        shortestPaths = sorted(shortestPaths)
        for i in range(0, len(shortestPaths)):
            path = shortestPaths[i].path
            if i == 0:  # shortest path gets bolder edges
                for j in range(len(path) - 1):
                    G[path[j]][path[j + 1]]['penwidth'] = 6.0
                    G[path[j]][path[j + 1]]['style'] = 'solid'
                    G.nodes[path[j]]['penwidth'] = 6.0
                    G.nodes[path[j + 1]]['penwidth'] = 6.0
                    G[path[j]][path[j + 1]]['color'] = '#778899FF'
                    # make the nodes look opaque if they are connected to the source
                    if len(G.nodes[path[j]]['fillcolor']) != 9:
                        G.nodes[path[j]]['fillcolor'] += 'FF'
                        G.nodes[path[j]]['color'] = '#708090FF'
                    if len(G.nodes[path[j + 1]]['fillcolor']) != 9:
                        G.nodes[path[j + 1]]['fillcolor'] += 'FF'
                        G.nodes[path[j + 1]]['color'] = '#708090FF'
            else:
                for j in range(len(path) - 1):
                    G[path[j]][path[j + 1]]['penwidth'] = 6.0
                    G[path[j]][path[j + 1]]['style'] = 'solid'
                    G.nodes[path[j]]['penwidth'] = 6.0
                    G.nodes[path[j + 1]]['penwidth'] = 6.0
                    if G[path[j]][path[j + 1]]['color'] != '#778899FF':
                        G[path[j]][path[j + 1]]['color'] = '#7788997F'
                    # make the nodes look opaque if they are connected to the source
                    if len(G.nodes[path[j]]['fillcolor']) != 9:
                        G.nodes[path[j]]['fillcolor'] += '7F'
                        G.nodes[path[j]]['color'] = '#7080907F'
                    if len(G.nodes[path[j + 1]]['fillcolor']) != 9:
                        G.nodes[path[j + 1]]['fillcolor'] += '7F'
                        G.nodes[path[j + 1]]['color'] = '#7080907F'
            shortestPaths[i].set_id("1" + letters[i])
        br = Branch(1, shortestPaths[0].path[-1])
        for pt in shortestPaths:
            br.add_path(pt)
        return [br]
    else:  # no paths found
        raise RuntimeError("No paths to target found.")



import os
import sys
import unittest
import warnings
import pyemap
from math import isclose

class SingleChainParams(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.my_emap = pyemap.parse(os.path.join(sys.path[0],"pyemap/tests/test_pdbs/1u3d.pdb")) 

    def test_eta_moities(self):
        assert len(self.my_emap.eta_moieties) >= 3

    def test_aromatic_amino_acid_options(self):
        pyemap.process(self.my_emap)
        ##default, tryptophan and tyrosine on
        resnames=[]
        for residue in self.my_emap.residues.values():
            resnames.append(residue.resname)
        assert "TYR" in resnames
        assert "TRP" in resnames
        assert "PHE" not in resnames
        assert "HIS" not in resnames
        #phenylalanine on
        pyemap.process(self.my_emap,phe=True)
        resnames=[]
        for residue in self.my_emap.residues.values():
            resnames.append(residue.resname)
        assert "TYR" in resnames
        assert "TRP" in resnames
        assert "PHE" in resnames
        assert "HIS" not in resnames
        #histidine on
        resnames=[]
        pyemap.process(self.my_emap,his=True)
        for residue in self.my_emap.residues.values():
            resnames.append(residue.resname)
        assert "TYR" in resnames
        assert "TRP" in resnames
        assert "PHE" not in resnames
        assert "HIS" in resnames
        #tyrosine off
        resnames=[]
        pyemap.process(self.my_emap,tyr=False)
        for residue in self.my_emap.residues.values():
            resnames.append(residue.resname)
        assert "TYR" not in resnames
        assert "TRP" in resnames
        assert "PHE" not in resnames
        assert "HIS" not in resnames
        #tryptophan off
        resnames=[]
        pyemap.process(self.my_emap,trp=False)
        for residue in self.my_emap.residues.values():
            resnames.append(residue.resname)
        assert "TYR" in resnames
        assert "TRP" not in resnames
        assert "PHE" not in resnames
        assert "HIS" not in resnames


    def test_distance_options(self):   
        pyemap.process(self.my_emap)
        com_weight = self.my_emap.init_graph["W385(A)"]["Y53(A)"]['weight']
        assert isclose(com_weight,6.18,abs_tol=1e-2)
        pyemap.process(self.my_emap,dist_def=1)
        closest_atom_weight = self.my_emap.init_graph["W385(A)"]["Y53(A)"]['weight']
        assert isclose(closest_atom_weight,3.97,abs_tol=1e-2)

    def test_choose_eta_moieties(self):        
        pyemap.process(self.my_emap)
        eta_moieties_resnames = []
        for res in self.my_emap.eta_moieties:
            eta_moieties_resnames.append(res)
        #default
        resnames=[]
        for residue in self.my_emap.residues.values():
            resnames.append(residue.resname)
        assert all(elem in resnames for elem in eta_moieties_resnames)
        #deselected eta moiety
        pyemap.process(self.my_emap,eta_moieties=eta_moieties_resnames[:-1])
        resnames=[]
        for residue in self.my_emap.residues.values():
            resnames.append(residue.resname)
        assert all(elem in resnames for elem in eta_moieties_resnames[:-1])
        assert eta_moieties_resnames[-1] not in resnames
    
    def test_custom_atom_range(self):
         #custom residues
        pyemap.process(self.my_emap,eta_moieties=[], custom = "(134-140),(109-117)")
        resnames=[]
        for residue in self.my_emap.residues.values():
            resnames.append(residue.resname)
        assert all(elem in resnames for elem in ["CUST-1","CUST-2"])
        #example for bad custom residues, shoud raise exception
        try:
            pyemap.process(self.my_emap, custom = "(28-41)")
            assert False
        except Exception as e:
            assert True

    def test_graph_parameters(self):
        pyemap.process(self.my_emap,num_st_dev_edges=5,percent_edges=3)
        g1 = self.my_emap.init_graph
        # check st_dev_edges works
        pyemap.process(self.my_emap,num_st_dev_edges=1,percent_edges=3)
        g2 = self.my_emap.init_graph
        assert g1.edges() != g2.edges()
        #check distance cutoff works    
        pyemap.process(self.my_emap,num_st_dev_edges=5,percent_edges=3,distance_cutoff=30)
        g3 = self.my_emap.init_graph
        assert g1.edges() != g3.edges()
        #check percent edges works
        pyemap.process(self.my_emap,num_st_dev_edges=5,percent_edges=1,distance_cutoff=30)
        g4 = self.my_emap.init_graph
        assert g1.edges() != g4.edges()

    def test_penalty_function_parameters(self):  
        #alpha
        pyemap.process(self.my_emap,coef_alpha=0.5)
        alpha_weight = self.my_emap.init_graph["W385(A)"]["Y53(A)"]['weight']
        assert isclose(alpha_weight,6.48,abs_tol=1e-2)
        #Roffset
        pyemap.process(self.my_emap,r_offset=1)
        alpha_weight = self.my_emap.init_graph["W385(A)"]["Y53(A)"]['weight']
        assert isclose(alpha_weight,5.18,abs_tol=1e-2)
        #Roffset
        pyemap.process(self.my_emap,exp_beta=5)
        beta_weight = self.my_emap.init_graph["W385(A)"]["Y53(A)"]['weight']
        assert isclose(beta_weight,13.44,abs_tol=1e-2)


class MultiChainParams(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.my_emap = pyemap.parse(os.path.join(sys.path[0],"pyemap/tests/test_pdbs/2oal.pdb")) 

    def test_chains(self):
        #default all chains on
        pyemap.process(self.my_emap)
        labels=[]
        for residue in self.my_emap.residues.values():
            labels.append(residue.node_label)
        assert all(elem in labels for elem in ["W18(A)" and 'Y298(B)'])
        #B chain off
        pyemap.process(self.my_emap,chains=["A"])
        labels=[]
        for residue in self.my_emap.residues.values():
            labels.append(residue.node_label)
        assert "W18(A)" in labels
        assert 'Y298(B)' not in labels



# PyeMap: A python package for automatic identification of electron and hole transfer pathways in proteins.
# Copyright(C) 2017-2020 Ruslan Tazhigulov, James Gayvert, Ksenia Bravaya (Boston University, USA)
"""Functions for identifying electron transfer active moieties, and constructing customized Bio.PDB.Residue objects.
"""
import networkx as nx
import numpy as np
from .data import SB_means,SB_std_dev, clusters
from .structures import cleanup_bonding, remove_side_chains

def is_pi_bonded(cur_atom, next_atom):
    """Determines whether two atoms are pi bonded based on experimental bond lengths.

    Parameters
    ----------
    cur_atom: BioPython Atom object
    next_atom: BioPython Atom object
        The two atoms being considered

    Returns
    -------
    True: if within 3 standard deviations of equilibrium pi bond length
    False: Otherwise

    References
    ----------
    Haynes, W. M. (Ed.). (2014). CRC Handbook of Chemistry and Physics. CRC Press.
        Reference for equilibrium pi bond lengths and standard deviations

    """
    x1 = cur_atom.coord[0]
    y1 = cur_atom.coord[1]
    z1 = cur_atom.coord[2]
    x2 = next_atom.coord[0]
    y2 = next_atom.coord[1]
    z2 = next_atom.coord[2]
    v1 = np.array((x1, y1, z1))
    v2 = np.array((x2, y2, z2))
    bond = str(cur_atom.element.upper()) + str(next_atom.element.upper())
    cutoff = SB_means.get(bond)
    if cutoff:
        cutoff -= 2 * SB_std_dev.get(bond)
        return dist(v1, v2) <= cutoff
    else:
        return False

def dist(x, y):
    """Returns Euclidean distance between two 3 dimensional points.

    Parameters
    ----------
    x: 3D numpy array
        Point 1
    y: 3D numpy array
        Point 2

    Returns
    -------
    float:
        Euclidean distance between two points

    """
    return np.sqrt(np.sum((x - y)**2))

def find_conjugated_systems(atoms, res_names):
    """Finds conjugated systems within a BioPython residue object, and returns them as individual customized BioPython
    Residue objects.

    Parameters
    ---------
    atoms: array-like
        List of atoms in the residue
    res_names: arary-like
        List of already used names for custom residues

    Returns
    -------
    custom_res_list: array-like
        List of customized :class:`Bio.PDB.Residue.Residue`
    """
    # first let's create the chemical graph structure
    # only these elements will be considered in our search
    arom_atoms = ['O', 'P', 'N', 'C', 'S']
    cust_graph = nx.Graph()
    for i in range(len(atoms)):
        for k in range(i, len(atoms)):
            if (not i == k) and is_pi_bonded(atoms[i], atoms[k]):
                if atoms[i].element in arom_atoms and atoms[k].element in arom_atoms:
                    cust_graph.add_edge(i, k)
                    cust_graph.nodes[i]["element"] = atoms[i].element
                    cust_graph.nodes[i]["coords"] = atoms[i].coord
                    cust_graph.nodes[k]["element"] = atoms[k].element
                    cust_graph.nodes[k]["coords"] = atoms[k].coord
    # now we have a forest, let's get each individual tree (only take cycles or bigger than 10)
    subgraphs = []
    all_subgraphs = (cust_graph.subgraph(c)
                     for c in nx.connected_components(cust_graph))
    for sub_g in all_subgraphs:
        graph = sub_g.copy()
        if nx.cycle_basis(graph):
            cleanup_bonding(graph)
            remove_side_chains(graph)
            subgraphs.append(graph)
        elif len(graph.nodes()) >= 10:
            subgraphs.append(graph)
    custom_res_list = []
    count = 1

    # iterate over all of the subgraphs
    for graph in subgraphs:
        res_name = str(atoms[0].parent.get_resname()) + str(
            atoms[0].get_full_id()[3][1]) + "(" + str(
                atoms[0].get_full_id()[2]) + ")"
        res_name = res_name.strip()
        if len(subgraphs) > 1:
            res_name += '-' + str(count)
            while res_name in res_names:
                count += 1
                res_name = res_name.split('-')[0] + '-' + str(count)
        atm_list = []
        for node in graph.nodes():
            atm_list.append(atoms[node])
        custom_res_list.append(create_custom_residue(atm_list, res_name))
        res_names.append(res_name)

    return custom_res_list

def create_custom_residue(atm_list, res_name):
    """Generates customized BioPython residue object corresponding to an ETA moiety.

    Parameters
    ----------
    atm_list: array-like
        List of atoms to be included
    res_name: str
        Name of custom residue

    Returns
    -------
    custom_residue: BioPython Residue Object
        customized BioPython residue object corresponding to a conjugated system

    """
    # copy residue, assign resname
    atm_list[0].parent.get_full_id()  # initialize instance variables
    custom_res = atm_list[0].get_parent().copy()
    custom_res.resname = res_name
    custom_res.node_label = res_name

    # get serial numbers of aromatic atoms
    atm_serial_list = []
    for atm in atm_list:
        atm_serial_list.append(atm.serial_number)

    # remove atoms not in aromatic ring by serial number
    atm_list = list(custom_res.get_atoms())
    for k in range(0, len(atm_list)):
        if atm_list[k].serial_number not in atm_serial_list:
            custom_res.detach_child(atm_list[k].id)

    return custom_res

def process_custom_residues(non_standard_residue_list):
    """Identifies and returns customized Bio.PDB.Residue objects corresponding to electron transfer active moieties.

    Parameters
    ---------
    non_standard_residue_list: list of :class:`Bio.PDB.Residue.Residue`
        List of non-protein residues in the structure

    Returns
    ------
    custom_res: list of :class:`Bio.PDB.Residue.Residue`
        List of customized BioPython residue objects corresponding to electron transfer active moieties that are not part of
        standard protein residues

    """
    res_names = []
    custom_res = []

    for residue in non_standard_residue_list:
        atm_list = list(residue.get_atoms())
        conj_systems = find_conjugated_systems(
            atm_list, res_names)
        if conj_systems:
            for system in conj_systems:
                res_names.append(system.resname)
            custom_res += conj_systems

    # Clusters
    for residue in non_standard_residue_list:
        if residue.resname in clusters:
            atm_list = list(residue.get_atoms())
            res_name = str(atm_list[0].parent.get_resname()) + str(
                atm_list[0].get_full_id()[3][1]) + "(" + str(
                    atm_list[0].get_full_id()[2]) + ")"
            custom_res.append(create_custom_residue(atm_list, res_name))
    return custom_res



# PyeMap: A python package for automatic identification of electron and hole transfer pathways in proteins.
# Copyright(C) 2017-2020 Ruslan Tazhigulov, James Gayvert, Ksenia Bravaya (Boston University, USA)
import numpy as np
from .shortest_paths import yens_shortest_paths, dijkstras_shortest_paths

def _finish_graph(G, original_shape_start, source):
    """Draws the graph with the shortest pathways highlighted.

    Parameters
    ----------
    G: NetworkX graph object
        Residue graph
    source: str
        name of source node
    original_shape_start: str
        shape of source node

    """
    G.nodes[source]['fillcolor'] = '#FFD700FF'
    G.nodes[source]['penwidth'] = 6.0
    G.nodes[source]['shape'] = original_shape_start
    # if not is not involved in pathways, make it less opaque on the graph.
    # change font color to slate gray, and change transparency of edges
    for name_node in G.nodes():
        if len(G.nodes[name_node]['fillcolor']) != 9:
            G.nodes[name_node]['fillcolor'] += '40'
            G.nodes[name_node]['fontcolor'] = '#708090'
    for edge in G.edges():
        name_node1, name_node2 = edge[0], edge[1]
        if G[name_node1][name_node2]['style'] == 'dashed':
            G[name_node1][name_node2]['color'] = '#7788994F'
    # draw graph


def find_paths(emap, source, target=None, max_paths=10):
    """Function which calculates pathways from source to target or surface exposed residues.

    Performs shortest path analysis on source and (optionally) target residues.

    Parameters
    ---------
    emap: :class:`~pyemap.emap` 
        Object for storing state of emap analysis.
    source: str
        source node for analysis
    target: str, optional
        target node for analysis
    max_paths: int, optional
        maximum number of paths to search for in yen's algorithm
    """
    # read in graph from file
    emap._reset_paths()
    G = emap.init_graph.copy()
    for u, v, d in G.edges(data=True):
        d['weight'] = np.float64(d['weight'])
    # process source and target
    source = source.strip()
    original_shape_start = G.nodes[source]['shape']
    G.nodes[source]['shape'] = 'oval'
    if target:
        target = target.strip()
        branches = yens_shortest_paths(G, source, target, max_paths=max_paths)
        # color target node blue
        G.nodes[target]['fillcolor'] = '#40e0d0FF'
        G.nodes[target]['penwidth'] = 6.0
    else:
        surface_exposed = emap.get_surface_exposed_residues()
        if source in surface_exposed:
            surface_exposed.remove(source)
        branches = dijkstras_shortest_paths(G, source, surface_exposed)
    _finish_graph(G, original_shape_start, source)
    emap._store_paths_graph(G)
    if target:
        emap._store_paths(branches, yens=True)
    else:
        emap._store_paths(branches)



import os
import sys
import unittest
import warnings
import pyemap

def test_parse():
    #from PDB file
    my_emap = pyemap.parse(os.path.join(sys.path[0],"pyemap/tests/test_pdbs/1u3d.pdb"))  
    assert len(list(my_emap.structure.get_residues()))==577 
    #from CIF file
    my_emap = pyemap.parse(os.path.join(sys.path[0],"pyemap/tests/test_pdbs/1u3d.cif"))  
    assert len(list(my_emap.structure.get_residues()))==577 


def test_fetch():
    #fetch from pdb database
    my_emap = pyemap.fetch_and_parse("1u3d")
    os.remove("1u3d.pdb")
    assert len(list(my_emap.structure.get_residues()))==577



# PyeMap: A python package for automatic identification of electron and hole transfer pathways in proteins.
# Copyright(C) 2017-2020 Ruslan Tazhigulov, James Gayvert, Ksenia Bravaya (Boston University, USA)

# Add imports here
from .parser import fetch_and_parse,parse
from .process_data import process
from .pathway_analysis import find_paths
from .emap import emap
from .shortest_paths import ShortestPath, Branch




"""PyeMap: A python package for automatic identification of electron and hole transfer pathways in proteins.

PyeMap (pronounced Pie-Map) is a python package for automatic identification and visualization of electron and hole transfer pathways in proteins. 
PyeMap analysis happens in 3 steps. The first step is parsing a PDB or CIF file provided by the user or fetched from the RCSB database. 
The next step is constructing the graph theory model of the protein crystal structure. Finally, the shortest paths between a 
specified electron/hole source to the surface or to a specified electron/hole acceptor can be calculated.

All PyeMap wheels distributed on PyPI are 3-BSD licensed.

GitHub: https://github.com/gayverjr/pyemap

Documentation: https://pyemap.readthedocs.io/en/latest/
"""
import sys
from setuptools import setup, find_packages

DOCLINES = (__doc__ or '').split("\n")

# from https://github.com/pytest-dev/pytest-runner#conditional-requirement
needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []


setup(
    # Self-descriptive entries which should always be present
    name='pyemap',
    author='James Gayvert',
    author_email='jrg444@gmail.com',
    description= DOCLINES[0],
    long_description="\n".join(DOCLINES[2:]),
    version="1.0.4",
    license='BSD-3-Clause',

    # Which Python importable modules should be included when your package is installed
    # Handled automatically by setuptools. Use 'exclude' to prevent some specific
    # subpackage(s) from being added, if needed
    packages=find_packages(),

    # Optional include package data to ship with your package
    # Customize MANIFEST.in if the general case does not suit your needs
    # Comment out this line to prevent the files from being packaged with your software
    include_package_data=True,

    # Allows `setup.py test` to work correctly with pytest
    setup_requires=[] + pytest_runner,

    # Additional entries you may want simply uncomment the lines you want and fill in the data
    # url='http://www.my_package.com',  # Website
    install_requires=['numpy',
    				  'networkx',
    				  'biopython',
    			      'scipy',
    				  'pillow',
                      'reportlab',
                      'pygraphviz',
                      'svglib']
    				   # Required packages, pulls from pip if needed; do not use for Conda deployment
    # platforms=['Linux',
    #            'Mac OS-X',
    #            'Unix',
    #            'Windows'],            # Valid platforms your code works on, adjust to your flavor
    #python_requires=">=3.7"         # Python version restrictions

    # Manual control if final package is compressible or not, set False to prevent the .egg from being made
    # zip_safe=False,

)



# PyeMap: A python package for automatic identification of electron and hole transfer pathways in proteins.
# Copyright(C) 2017-2020 Ruslan Tazhigulov, James Gayvert, Ksenia Bravaya (Boston University, USA)
"""Parser used to parse PDB and MMCIF files.

Constructs an emap object containing parsed Bio.PDB.Structure and a list of customized Bio.PDB.Residue objects
corresponding to automatically identified electron transfer moieties.
"""
from Bio.PDB import PDBIO, FastMMCIFParser, PDBParser
from .custom_residues import process_custom_residues
from .data import res_name_to_char
from .emap import emap
import os

def download_pdb(pdbcode, datadir, downloadurl="https://files.rcsb.org/download/"):
    """
    Downloads a PDB file from the Internet and saves it in a data directory.
    :param pdbcode: The standard PDB ID e.g. '3ICB' or '3icb'
    :param datadir: The directory where the downloaded file will be saved
    :param downloadurl: The base PDB download URL, cf.
        `https://www.rcsb.org/pages/download/http#structures` for details
    :return: the full path to the downloaded PDB file or None if something went wrong
    """
    import urllib
    pdbfn = pdbcode + ".pdb"
    url = downloadurl + pdbfn
    outfnm = os.path.join(datadir, pdbfn)
    try:
        urllib.request.urlretrieve(url, outfnm)
        return outfnm
    except Exception as err:
        print(str(err), file=sys.stderr)
        return None

def fetch_and_parse(pdb_id, dest="", quiet=False):
    '''Fetches pdb from database and parses the file.

    Parameters
    ----------
    pdb_id: str
        RCSB PDB ID
    dest: str, optional
        Full path to where file should be saved
    quiet: bool, optional
        Supresses output when set to true

    Returns
    -------
    emap: :class:`~pyemap.emap`
        emap object ready for processing.
    '''
    if not dest:
        dest = os.getcwd()
    if not quiet:
        print("Fetching PDB " + pdb_id + " from RSCB Database...")
    outfnm = download_pdb(pdb_id,dest)
    print(outfnm)
    if not quiet:
        print("Success!")
    return parse(outfnm, quiet)

def parse(filename, quiet=False):
    '''Parses pdb file and returns emap object.
    
    Parameters
    ----------
    filename: str
        Full path to file which needs to be parsed
    quiet: bool, optional
        Supresses output when set to true
    
    Returns
    -------
    my_emap: :class:`~pyemap.emap`
        emap object reading for parsing
    '''
    print(filename)
    try:
        os.listdir()
        parser = PDBParser()
        structure = parser.get_structure("protein", filename)
    except Exception as e:
        parser = FastMMCIFParser()
        structure = parser.get_structure("protein", filename)
        io = PDBIO()
        fn = filename[:-4] + ".pdb"
        io.set_structure(structure)
        io.save(fn)
        parser = PDBParser()
        structure = parser.get_structure("protein", fn)
    chain_list = []
    num_models = 0
    for model in structure.get_models():
        num_models += 1
    if num_models < 1:
        raise RuntimeError("Unable to parse file.")
    for chain in structure[0].get_chains():
        chain_list.append(chain.id)
    non_standard_residue_list = []
    for res in structure[0].get_residues():
        if res.resname not in res_name_to_char:
            res.get_full_id()
            arom_res = res.copy()
            non_standard_residue_list.append(arom_res)
    custom_residue_list = process_custom_residues(non_standard_residue_list)
    if not quiet:
        print("Identified " + str(len(custom_residue_list)) + " non-protein ET active moieties.")
    my_emap = emap(filename, structure, custom_residue_list, chain_list)
    return my_emap




# PyeMap: A python package for automatic identification of electron and hole transfer pathways in proteins.
# Copyright(C) 2017-2020 Ruslan Tazhigulov, James Gayvert, Ksenia Bravaya (Boston University, USA)
from heapq import heappop, heappush
import networkx as nx
import numpy as np
from .data import SB_means,SB_std_dev
from collections import defaultdict


def buildSmarts(graph, cur, prev):
    '''DFS algorithm to generate Smarts string.

    Parameters
    ----------
    graph: :class:`networkx.Graph`
    atoms: array-like
        List of atoms in structure
    cur,prev: int 
        Atom indices

    References
    ----------
    Varnek, A. Tutorials in Chemoinformatics; John Wiley & Sons, Inc.: Hoboken, NJ, 2017.
    '''
    visited.add(cur)
    seq = ''
    seq += graph.nodes[cur]["element"].lower()
    for d in closingClosures[cur]:
        seq += d
        heappush(digits, d[-1])
    for a in openingClosures[cur]:
        d = str(heappop(digits))
        seq += d
        closingClosures[a].append(d)
    branches = []
    neighbors = list(graph.neighbors(cur))
    if prev in neighbors:
        neighbors.remove(prev)
    for neighbor in neighbors:
        if neighbor not in visited:
            branches.append(buildSmarts(graph, neighbor, cur))
    for branch in branches[:-1]:
        seq += "(" + branch + ")"
    if len(branches) > 0:
        seq += branches[-1]
    return seq


def getClosures(graph, cur, prev):
    '''DFS algorithm to generate Smarts string.

    graph: :class:`networkx.Graph`
        Chemical graph
    cur,prev: int
        Atom indices

    References
    ----------
    Varnek, A. Tutorials in Chemoinformatics; John Wiley & Sons, Inc.: Hoboken, NJ, 2017.

    '''
    ancestor.add(cur)
    visited.add(cur)
    neighbors = list(graph.neighbors(cur))
    if prev in neighbors:
        neighbors.remove(prev)
    for neighbor in neighbors:
        if neighbor in ancestor:
            openingClosures[neighbor].append(cur)
        elif neighbor not in visited:
            getClosures(graph, neighbor, cur)
    ancestor.remove(cur)


def getSimpleSmarts(graph):
    """DFS algorithm to generate Smarts string.

    Parameters
    ----------
    graph: :class:`networkx.Graph`
        Chemical graph

    References
    ----------
    Varnek, A. Tutorials in Chemoinformatics; John Wiley & Sons, Inc.: Hoboken, NJ, 2017.

    """
    root = list(graph.nodes())[0]
    global visited, ancestor, openingClosures, closingClosures, digits
    visited = set()
    ancestor = set()
    openingClosures = defaultdict(list)
    getClosures(graph, root, None)
    closingClosures = defaultdict(list)
    digits = [str(x) for x in range(1, 10)]
    visited = set()
    return buildSmarts(graph, root, None)


def is_part_of_cycle(node, res_graph):
    for cycle in nx.cycle_basis(res_graph):
        if node in cycle:
            return True
    return False


def is_close(node, node2, res_graph):
    bond = str(res_graph.nodes[node]["element"].upper()) + str(res_graph.nodes[node2]["element"].upper())
    dist = np.sqrt(np.sum((np.array(res_graph.nodes[node]["coords"]) -\
     						np.array(res_graph.nodes[node2]["coords"]))**2))
    cutoff = SB_means.get(bond) + 3 * SB_std_dev.get(bond)
    return dist < cutoff


def cleanup_bonding(res_graph):
    '''Connects nodes that should be connected to fix broken aromaticity.
    
    Parameters
    -----------
    res_graph: :class:`networkx.Graph`
        residue graph
    '''
    for node in res_graph.nodes:
        if not is_part_of_cycle(node, res_graph) and len(list(res_graph.neighbors(node))) < 3:
            closest_neighbor = []
            min_dist = 10000
            for node2 in res_graph.nodes:
                if node2 != node and node2 not in res_graph.neighbors(node) and len(list(
                        res_graph.neighbors(node2))) < 3:
                    if is_close(node, node2, res_graph):
                        dist = np.sqrt(np.sum((np.array(res_graph.nodes[node]["coords"]) -\
                         					   np.array(res_graph.nodes[node2]["coords"]))**2))
                        if dist < min_dist:
                            min_dist = dist
                            closest_neighbor = node2
            if closest_neighbor:
                res_graph.add_edge(node, closest_neighbor)


def remove_atoms(prev, cur, remove_list, res_graph):
    if not is_part_of_cycle(cur, res_graph):
        for neighbor in res_graph.neighbors(cur):
            if neighbor != prev:
                remove_atoms(cur, neighbor, remove_list, res_graph)
        remove_list.append(cur)


def remove_side_chains(res_graph):
    ''' Removes non-aromatic sides chains on aromatic eta moieties.

    Parameters
    -----------
    res_graph: :class:`networkx.Graph`
        residue graph
    '''
    remove_list = []
    for node in res_graph.nodes:
        if not is_part_of_cycle(node,
                                res_graph) and res_graph.nodes[node]["element"] == "C" and node not in remove_list:
            remove_atoms(-1, node, remove_list, res_graph)
    for node in remove_list:
        res_graph.remove_node(node)
